unit UMainForm;

{$mode Delphi}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, SynEdit, Forms, Controls, Graphics, Dialogs,
  Menus, ActnList, StdCtrls, ComCtrls, UBrainFuck, UCoolTranslator, StrUtils,
  SpecializedList, UCompiler, Registry, URegistry, ULastOpenedList, Process;

const
  RegistryRoot = HKEY_CURRENT_USER;

type

  { TMainForm }

  TMainForm = class(TForm)
    ACompile: TAction;
    AAbout: TAction;
    ABreakpointToggle: TAction;
    ACompileAndRun: TAction;
    AViewCompilers: TAction;
    AGenerateNumber: TAction;
    AShrinkSource: TAction;
    AFormatSource: TAction;
    AOptions: TAction;
    AInterpretterStepOut: TAction;
    AInterpretterRunToCursor: TAction;
    AInterpretterStopOver: TAction;
    AInterpretterStepInto: TAction;
    AViewCompiled: TAction;
    AViewInterpretter: TAction;
    AProgramPause: TAction;
    AProgramStop: TAction;
    AHelp: TAction;
    AProgramRun: TAction;
    AProjectNew: TAction;
    AExit: TAction;
    AProjectSaveAs: TAction;
    AProjectClose: TAction;
    AProjectSave: TAction;
    AProjectOpen: TAction;
    ActionList1: TActionList;
    CoolTranslator1: TCoolTranslator;
    ImageList1: TImageList;
    MainMenu1: TMainMenu;
    MemoSource: TMemo;
    MenuItem1: TMenuItem;
    MenuItem10: TMenuItem;
    MenuItem11: TMenuItem;
    MenuItem12: TMenuItem;
    MenuItem13: TMenuItem;
    MenuItem14: TMenuItem;
    MenuItem15: TMenuItem;
    MenuItem16: TMenuItem;
    MenuItem17: TMenuItem;
    MenuItem18: TMenuItem;
    MenuItem19: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem20: TMenuItem;
    MenuItem21: TMenuItem;
    MenuItem22: TMenuItem;
    MenuItem23: TMenuItem;
    MenuItem24: TMenuItem;
    MenuItem25: TMenuItem;
    MenuItem26: TMenuItem;
    MenuItem27: TMenuItem;
    MenuItem28: TMenuItem;
    MenuItem29: TMenuItem;
    MenuItem30: TMenuItem;
    MenuItem31: TMenuItem;
    MenuItemOpenRecent: TMenuItem;
    MenuItemTarget: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    MenuItem9: TMenuItem;
    OpenDialog1: TOpenDialog;
    PopupMenuSource: TPopupMenu;
    SaveDialog1: TSaveDialog;
    StatusBar1: TStatusBar;
    TabControl1: TTabControl;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton10: TToolButton;
    ToolButton11: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    procedure ABreakpointToggleExecute(Sender: TObject);
    procedure ACompileAndRunExecute(Sender: TObject);
    procedure ACompileExecute(Sender: TObject);
    procedure AExitExecute(Sender: TObject);
    procedure AFormatSourceExecute(Sender: TObject);
    procedure AGenerateNumberExecute(Sender: TObject);
    procedure AOptionsExecute(Sender: TObject);
    procedure AProgramPauseExecute(Sender: TObject);
    procedure AProgramStopExecute(Sender: TObject);
    procedure AProjectCloseExecute(Sender: TObject);
    procedure AProjectNewExecute(Sender: TObject);
    procedure AProjectOpenExecute(Sender: TObject);
    procedure AProjectSaveAsExecute(Sender: TObject);
    procedure AProjectSaveExecute(Sender: TObject);
    procedure AProgramRunExecute(Sender: TObject);
    procedure AShrinkSourceExecute(Sender: TObject);
    procedure AViewCompiledExecute(Sender: TObject);
    procedure AViewCompilersExecute(Sender: TObject);
    procedure AViewInterpretterExecute(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure MemoSourceChange(Sender: TObject);
    procedure MemoSourceKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure MemoSourceKeyPress(Sender: TObject; var Key: char);
    procedure MemoSourceKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState
      );
    procedure MemoSourceMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    procedure AProjectOpenRecentExecute(Sender: TObject);
    procedure BrainFuckInterpreterChangeState(Sender: TObject);
    procedure MenuItemTargetClick(Sender: TObject);
    procedure ProjectOpen(FileName: string);
  public
    Modified: Boolean;
    CompilerIndex: Integer;
    ProjectFileName: string;
    BrainFuckCompiler: TBrainFuckCompiler;
    BrainFuckInterpreter: TBrainFuckInterpretter;
    BreakPoints: TListInteger;
    Compilers: TCompilerList;
    LastOpenedList: TLastOpenedList;
    OpenProjectOnStart: Boolean;
    procedure LoadFromRegistry(Root: HKEY; Key: string);
    procedure SaveToRegistry(Root: HKEY; Key: string);
    procedure UpdateInterface;
    procedure UpdateStatusBar;
    procedure UpdateTargetList;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.lfm}

uses
  UInterpretterForm, UApplicationInfo, UCompiledForm, UOptionsForm,
  UCompilerDelphi, UCompilerPHP, UCompilerC, UCompilersForm;

resourcestring
  SEnterNumber = 'Enter number';
  SNumberGeneration = 'Number generation';
  SProgramExited = 'Program exited';
  SCompilerNotFound = 'Compiler "%s" not found';
  SCompiledFileNotFound = 'Program compiled file "%s" not found';

{ TMainForm }

procedure TMainForm.FormShow(Sender: TObject);
begin
  if OpenProjectOnStart and (LastOpenedList.Count > 0) then
    ProjectOpen(LastOpenedList[0])
    else AProjectNew.Execute;
  UpdateInterface;
end;

procedure TMainForm.MemoSourceChange(Sender: TObject);
begin
  Modified := True;
  UpdateInterface;
end;

procedure TMainForm.MemoSourceKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
end;

procedure TMainForm.MemoSourceKeyPress(Sender: TObject; var Key: char);
begin
end;

procedure TMainForm.MemoSourceKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  UpdateStatusBar;
end;

procedure TMainForm.MemoSourceMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  UpdateStatusBar;
end;

procedure TMainForm.BrainFuckInterpreterChangeState(Sender: TObject);
begin
  UpdateInterface;
end;

procedure TMainForm.MenuItemTargetClick(Sender: TObject);
var
  I: Integer;
begin
  with TMenuItem(Sender) do begin
    CompilerIndex := MenuIndex;
    Checked := True;
    for I := 0 to Parent.Count - 1 do
      if Parent.Items[I].MenuIndex <> MenuIndex then Parent.Items[I].Checked := False
  end;
end;

procedure TMainForm.ProjectOpen(FileName: string);
begin
  MemoSource.Lines.LoadFromFile(UTF8Decode(FileName));
  LastOpenedList.AddItem(FileName);
  ProjectFileName := FileName;
  UpdateInterface;
  Modified := False;
end;

procedure TMainForm.LoadFromRegistry(Root: HKEY; Key: string);
begin
  with TRegistryEx.Create do
  try
    RootKey := Root;
    OpenKey(Key, True);
    OpenProjectOnStart := ReadBoolWithDefault('OpenProjectOnStart', True);
    BrainFuckInterpreter.CellSize := ReadIntegerWithDefault('CellSize', 256);
    BrainFuckInterpreter.MemorySize := ReadIntegerWithDefault('MemorySize', 30000);
    if ValueExists('LanguageCode') then
      CoolTranslator1.Language := CoolTranslator1.Languages.SearchByCode(ReadStringWithDefault('LanguageCode', ''))
      else CoolTranslator1.Language := CoolTranslator1.Languages.SearchByCode('');
    CompilerIndex := ReadIntegerWithDefault('SelectedCompiler', 0);
  finally
    Free;
  end;
  LastOpenedList.LoadFromRegistry(Root, Key);
  Compilers.LoadFromRegistry(Root, Key);
end;

procedure TMainForm.SaveToRegistry(Root: HKEY; Key: string);
begin
  Compilers.SaveToRegistry(Root, Key);
  LastOpenedList.SaveToRegistry(Root, Key);
  with TRegistryEx.Create do
  try
    RootKey := Root;
    OpenKey(Key, True);
    WriteBool('OpenProjectOnStart', OpenProjectOnStart);
    WriteInteger('CellSize', BrainFuckInterpreter.CellSize);
    WriteInteger('MemorySize', BrainFuckInterpreter.MemorySize);
    if Assigned(CoolTranslator1.Language) and (CoolTranslator1.Language.Code <> '') then
      WriteString('LanguageCode', CoolTranslator1.Language.Code)
      else DeleteValue('LanguageCode');
    WriteInteger('SelectedCompiler', CompilerIndex);
  finally
    Free;
  end;
end;

procedure TMainForm.UpdateInterface;
var
  Title: string;
begin
  AProjectSave.Enabled := Modified;
  Title := ApplicationInfo.Name;
  if ProjectFileName <> '' then Title := ProjectFileName + ' - ' + Title;
  if Modified then Title := Title + ' *';
  Caption := Title;
  MemoSource.Enabled := ProjectFileName <> '';
  AProjectClose.Enabled := ProjectFileName <> '';
  AProgramRun.Enabled := (ProjectFileName <> '') and (BrainFuckInterpreter.State = rsStopped);
  AProgramPause.Enabled := (ProjectFileName <> '') and (BrainFuckInterpreter.State = rsRunning);
  AProgramStop.Enabled := (ProjectFileName <> '') and (BrainFuckInterpreter.State <> rsStopped);
  ACompile.Enabled := ProjectFileName <> '';
  UpdateStatusBar;
  UpdateTargetList;
end;

procedure TMainForm.UpdateStatusBar;
begin
  StatusBar1.Panels[0].Text := IntToStr(MemoSource.CaretPos.X) + ', ' + IntToStr(MemoSource.CaretPos.Y);
end;

procedure TMainForm.UpdateTargetList;
var
  I: Integer;
  NewMenuItem: TMenuItem;
begin
  MenuItemTarget.Clear;
  for I := 0 to Compilers.Count - 1 do begin
    NewMenuItem := TMenuItem.Create(MenuItemTarget);
    NewMenuItem.Caption := TBrainFuckCompiler(Compilers[I]).Name;
    NewMenuItem.OnClick := MenuItemTargetClick;
    if I = CompilerIndex then NewMenuItem.Checked := True;
    MenuItemTarget.Add(NewMenuItem);
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  BreakPoints := TListInteger.Create;
  BrainFuckInterpreter := TBrainFuckInterpretter.Create;
  BrainFuckInterpreter.OnChangeState := BrainFuckInterpreterChangeState;
  BrainFuckCompiler := TBrainFuckCompiler.Create;
  Compilers := TCompilerList.Create;
  Compilers.Add(TBrainFuckCompilerDelphi.Create);
  Compilers.Add(TBrainFuckCompilerPHP.Create);
  Compilers.Add(TBrainFuckCompilerC.Create);
  UpdateTargetList;
  LastOpenedList := TLastOpenedList.Create;
  LastOpenedList.MenuItem := MenuItemOpenRecent;
  LastOpenedList.ClickAction := AProjectOpenRecentExecute;
  LoadFromRegistry(RegistryRoot, ApplicationInfo.RegistryKey);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  SaveToRegistry(RegistryRoot, ApplicationInfo.RegistryKey);
  LastOpenedList.Free;
  Compilers.Free;
  BrainFuckCompiler.Free;
  BrainFuckInterpreter.Free;
  BreakPoints.Free;
end;

procedure TMainForm.AProgramRunExecute(Sender: TObject);
begin
  InterpreterForm.LastStepCounter := 0;
  InterpreterForm.Show;
  BrainFuckInterpreter.Input := InterpreterForm.MemoInput.Lines.Text;
  BrainFuckInterpreter.Source := MemoSource.Text;
  BrainFuckInterpreter.Run;
end;

procedure TMainForm.AShrinkSourceExecute(Sender: TObject);
var
  Source: string;
  Pos: Integer;
  I: Integer;
begin
  Source := MemoSource.Text;
  Pos := 1;
  for I := 1 to Length(Source) do begin
    if Source[I] > ' ' then begin
      Source[Pos] := Source[I];
      Inc(Pos);
    end;
  end;
  SetLength(Source, Pos - 1);
  MemoSource.Text := Source;
end;

procedure TMainForm.AViewCompiledExecute(Sender: TObject);
begin
  CompiledForm.Show;
end;

procedure TMainForm.AViewCompilersExecute(Sender: TObject);
begin
  FormCompilers.Show;
end;

procedure TMainForm.AViewInterpretterExecute(Sender: TObject);
begin
  InterpreterForm.Show;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  if Modified then AProjectSaveAs.Execute;
end;

procedure TMainForm.AExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.AFormatSourceExecute(Sender: TObject);
var
  Source: string;
  NewSource: string;
  Indent: Integer;
  I: Integer;
const
  IndentText = '  ';
begin
  Source := MemoSource.Text;
  NewSource := '';
  Indent := 0;
  for I := 1 to Length(Source) do begin
    if Source[I] = '[' then begin
      NewSource := NewSource + LineEnding + DupeString(IndentText, Indent) + Source[I] + LineEnding ;
      Inc(Indent);
      NewSource := NewSource + DupeString(IndentText, Indent);
    end
    else if Source[I] = ']' then begin
      Dec(Indent);
      NewSource := NewSource + LineEnding + DupeString(IndentText, Indent) + Source[I] + LineEnding + DupeString(IndentText, Indent);
    end
    else if Ord(Source[I]) > $20 then
      NewSource := NewSource + Source[I];
  end;
  MemoSource.Text := NewSource;
end;

procedure TMainForm.AGenerateNumberExecute(Sender: TObject);
var
  NumberText: string;
  Number: Integer;
  I: Integer;
begin
  NumberText := InputBox(SNumberGeneration, SEnterNumber, '1');
  if TryStrToInt(NumberText, Number) then begin
    NumberText := '';
    for I := 0 to Number - 1 do
      NumberText := NumberText + '+';
    MemoSource.Text := MemoSource.Text + NumberText;
  end;
end;

procedure TMainForm.AOptionsExecute(Sender: TObject);
begin
  OptionsForm.LoadFromInterpretter(BrainFuckInterpreter);
  if OptionsForm.ShowModal = mrOK then begin
    OptionsForm.SaveToInterpretter(BrainFuckInterpreter);
  end;
end;

procedure TMainForm.ACompileExecute(Sender: TObject);
begin
  with TBrainFuckCompiler(Compilers[CompilerIndex]) do begin
    Optimization := coNormal;
    Source := MemoSource.Text;
    ProgramName := ExtractFileNameOnly(ProjectFileName);
    Compile;
    CompiledForm.MemoCompiled.Text := Output;
    CompiledForm.Show;
  end;
end;

procedure TMainForm.ABreakpointToggleExecute(Sender: TObject);
begin
  //BreakPoints.Add();
end;

procedure TMainForm.ACompileAndRunExecute(Sender: TObject);
var
  Process: TProcess;
  CompiledFile: string;
begin
  with TBrainFuckCompiler(Compilers[CompilerIndex]) do begin
    Optimization := coNormal;
    Source := MemoSource.Text;
    ProgramName := ExtractFileNameOnly(ProjectFileName);
    Compile;
    CompiledFile := ExtractFilePath(ProjectFileName) +
      'compiled' + DirectorySeparator + Name + DirectorySeparator +
      ExtractFileNameOnly(ProjectFileName) + SourceExtension;
    ForceDirectoriesUTF8(ExtractFilePath(CompiledFile));
    with TStringList.Create do
    try
      Text := Output;
      SaveToFile(CompiledFile);
    finally
      Free;
    end;
    if FileExistsUTF8(CompilerPath) then
    try
      Process := TProcess.Create(nil);
      Process.CommandLine := '"' + CompilerPath + '" "' + CompiledFile + '"';
      Process.Options := [poWaitOnExit];
      Process.Execute;
    finally
      Process.Free;
    end else raise Exception.Create(Format(SCompilerNotFound, [CompilerPath]));

    if CompiledExtension <> '' then begin
      CompiledFile := ExtractFilePath(CompiledFile) + ExtractFileNameOnly(CompiledFile) + CompiledExtension;
      if FileExistsUTF8(CompiledFile) then
      try
        Process := TProcess.Create(nil);
        Process.CommandLine := '"' + CompiledFile + '"';
        Process.Options := [poWaitOnExit];
        Process.Execute;
      finally
        Process.Free;
      end else raise Exception.Create(Format(SCompiledFileNotFound, [CompiledFile]));
    end;
  end;
end;

procedure TMainForm.AProgramPauseExecute(Sender: TObject);
begin
  MainForm.BrainFuckInterpreter.Pause
end;

procedure TMainForm.AProgramStopExecute(Sender: TObject);
begin
  MainForm.BrainFuckInterpreter.Stop;
end;

procedure TMainForm.AProjectCloseExecute(Sender: TObject);
begin
  Modified := False;
  ProjectFileName := '';
  MemoSource.Clear;
  UpdateInterface;
end;

procedure TMainForm.AProjectNewExecute(Sender: TObject);
begin
  ProjectFileName := 'Examples' + DirectorySeparator + 'Project.b';
  MemoSource.Clear;
  Modified := False;
  UpdateInterface;
end;

procedure TMainForm.AProjectOpenRecentExecute(Sender: TObject);
begin
  ProjectOpen(LastOpenedList[TMenuItem(Sender).MenuIndex]);
end;

procedure TMainForm.AProjectOpenExecute(Sender: TObject);
begin
  OpenDialog1.FileName := ProjectFileName;
  if OpenDialog1.Execute then begin
    ProjectOpen(OpenDialog1.FileName);
  end;
end;

procedure TMainForm.AProjectSaveAsExecute(Sender: TObject);
begin
  SaveDialog1.FileName := ProjectFileName;
  if SaveDialog1.Execute then begin
    MemoSource.Lines.SaveToFile(UTF8Decode(SaveDialog1.FileName));
    ProjectFileName := SaveDialog1.FileName;
    LastOpenedList.Add(SaveDialog1.FileName);
    Modified := False;
    UpdateInterface;
  end;
end;

procedure TMainForm.AProjectSaveExecute(Sender: TObject);
begin
  if Modified then AProjectSaveAs.Execute;
end;

end.

